# -----------------------------------------
# Sextant
# Copyright 2014, Ensoft Ltd.
# Author: James Harkin, using work from Patrick Stevens and James Harkin
# -----------------------------------------
# Invokes Sextant and argparse

from __future__ import absolute_import, print_function

import io
import sys
import socket
import logging
import logging.config
import argparse
import requests
import contextlib
import subprocess
import getpass

try:
    from urllib import parse
except ImportError:  # fall back to Python 2
    import urlparse as parse

from . import query
from . import db_api
from . import update_db
from . import environment
from . import sshmanager

config = environment.load_config()



def _displayable_url(args):
    """
    Return the URL specified by the user for Sextant to look at.

    This is needed because we may be using SSH (that is, Sextant sees a
    localhost port for the Neo4j server, which ssh is forwarding somewhere
    else); in this case, we want to display the URL that the user thinks
    Sextant is using, rather than the URL it's actually using.

    :param args: the argparse object Sextant was invoked with
    :return: the URL the Sextant invoker expects Sextant to be using

    """
    try:
        if args.display_neo4j:
            return args.display_neo4j
    except AttributeError:
        return args.remote_neo4j

    return args.remote_neo4j


# Beginning of functions which handle the actual invocation of Sextant

def _start_web(connection, args):
    # Don't import at top level - makes twisted dependency semi-optional,
    # allowing non-web functionality to work with Python 3.
    if sys.version_info[0] == 2:
        from .web import server
        server.render_timeout = config.render_timeout
    else:  # twisted won't be available - Python 2 required
        logging.error('Web server must be run on Python 2.')
        return
    logging.info("Serving site on port {}".format(args.port))

    # server is .web.server, imported a couple of lines ago
    server.serve_site(connection, args.port)


def _audit(connection, args):
    try:
        audited = query.audit(connection)
    except requests.exceptions.ConnectionError as e:
        msg = 'Connection error to server {url}: {exception}'
        logging.error(msg.format(url=_displayable_url(args), exception=e))
        return

    if not audited:
        location = _displayable_url(args)
        logging.warning('No programs on database at {}.'.format(location))
    else:
        # print data about all the programs in the database in a formatted table
        titles = ("Name", "#Func", "Uploader", "User-ID", "Upload Date")
        colminlens = (len(entry) for entry in titles)
        # maximum lengths to avoid one entry from throwing the whole table
        # date format is <YYYY-MM-DD HH:MM:SS> = 19 characters
        COLMAXLENS = (25, 6, 25, 10, 19)

        # make a table of the strings of each data entry we will display
        text = [map(str, (p.program_name, p.number_of_funcs,
                          p.uploader, p.uploader_id, p.date))
                for p in audited]

        # and a table of the corresponding lengths for every text entry
        textlens = ((len(entry) for entry in pentry) for pentry in text)

        # the maxes of the column maxes of the textlens table and the
        # lengths of the title row words give the column widths
        textcollens = map(max, zip(*textlens))
        # the lambda is exactly equivalent to calling
        # min(max(colminlens, textcollens), COLMAXLENS),
        # eg for x < low: sorted((low, x, high))[1] = (x, low, high)[1] = low
        collens = map(lambda l: sorted(l)[1], zip(colminlens, textcollens, COLMAXLENS))

        # insert the column widths into the format string
        st = ('{{0:{0}}}  {{1:>{1}}}  {{2:{2}}}  {{3:{3}}}  '
              '{{4:{4}}}'.format(*collens))

        # write the title line so we know how long to make the separator line
        titleline = st.format(*titles)
        sepline = '-' * len(titleline)

        # output the formatted table
        print(titleline)
        print(sepline)
        print('\n'.join(st.format(*pentry) for pentry in text))


def _add_program(connection, args):
    try:
        alternative_name = args.name_in_db[0]
    except TypeError:
        alternative_name = None

    not_object_file = args.not_object_file
    # the default is "yes, this is an object file" if not-object-file was
    # unsupplied

    try:
        update_db.upload_program(connection, 
                                 getpass.getuser(),
                                 args.input_file,
                                 alternative_name,
                                 not_object_file)
    except requests.exceptions.ConnectionError as e:
        msg = 'Connection error to server {}: {}'
        logging.error(msg.format(_displayable_url(args), e))

    except IOError as e:
        # in case of Python 2, where FileNotFoundError is undefined
        # note: ConnectionError subclasses IOError, so must come first
        logging.error('Input file {} was not found.'.format(args.input_file[0]))
        logging.error(e)
        logging.debug(e, exc_info=True)
    except (ValueError, sshmanager.SSHConnectionError) as e:
        logging.error(e)


def _delete_program(connection, args):
    update_db.delete_program(connection, args.program_name)


def _make_query(connection, args):
    arg1 = None
    arg2 = None
    try:
        arg1 = args.funcs[0]
        arg2 = args.funcs[1]
    except TypeError:
        pass
    except IndexError:
        pass

    try:
        program_name = args.program[0]
    except TypeError:
        program_name = None

    try:
        suppress_common = args.suppress_common[0]
    except TypeError:
        suppress_common = False

    query.query(remote_neo4j=args.remote_neo4j,
                display_neo4j=_displayable_url(args),
                input_query=args.query,
                program_name=program_name,
                argument_1=arg1,
                argument_2=arg2,
                suppress_common=suppress_common)

# End of functions which invoke Sextant

def parse_arguments():
    """
    Parses the command-line arguments to Sextant.

    The resulting object, result, has a .func property, which is a method to be
    called with result as its only parameter. This .func method runs whichever
    of Sextant's functionality is appropriate.
    :return: namespace summarising the arguments

    """

    ap = argparse.ArgumentParser(prog='sextant',
                                 usage='sextant', 
                                 description="Invoke part of the SEXTANT program")
    subparsers = ap.add_subparsers(title="subcommands")

    #set what will be defined as a "common function"
    db_api.set_common_cutoff(config.common_cutoff)

    parsers = dict()

    # add each subparser in turn to the parsers dictionary

    parsers['add'] = subparsers.add_parser('add-program',
                                           help="add a program to the database")
    parsers['add'].add_argument('input_file', metavar="FILE_NAME",
                                help="name of file to be put into database",
                                type=str)
    parsers['add'].add_argument('--name-in-db', metavar="PROGRAM_NAME",
                                help="string to store this program under", type=str,
                                nargs=1)
    parsers['add'].add_argument('--not-object-file',
                                help='default False, if the input file is an '
                                     'object to be disassembled',
                                action='store_true')

    parsers['delete'] = subparsers.add_parser('delete-program',
                                              help="delete a program from the database")
    parsers['delete'].add_argument('program_name', metavar="PROG_NAME",
                                   help="name of program as stored in the database",
                                   type=str)

    parsers['query'] = subparsers.add_parser('query',
                                             help="make a query of the database")
    parsers['query'].add_argument('query', metavar="QUERY",
                                  help="functions-calling, functions-called-by, "
                                       "all-call-paths, whole-program, "
                                       "shortest-call-path, programs or "
                                       "functions; if the latter, "
                                       "supply argument --program",
                                  type=str)
    parsers['query'].add_argument('--program', metavar="PROG_NAME",
                                  help="name of program as stored in the database; "
                                       "required for all queries except 'programs'",
                                  type=str, nargs=1)
    parsers['query'].add_argument('--funcs', metavar='FUNCS',
                                  help='functions to pass to the query',
                                  type=str, nargs='+')
    parsers['query'].add_argument('--suppress-common', metavar='BOOL',
                                  help='suppress commonly called functions (True or False)',
                                  type=str, nargs=1)

    parsers['web'] = subparsers.add_parser('web', help="start the web server")
    parsers['web'].add_argument('--port', metavar='PORT', type=int,
                                help='port on which to serve Sextant Web',
                                default=config.web_port)

    parsers['audit'] = subparsers.add_parser('audit', help='view usage of Sextant')

    for key in parsers:
        parsers[key].add_argument('--remote-neo4j', metavar="URL",
                                  help="URL of neo4j server", type=str,
                                  default=config.remote_neo4j)
        parsers[key].add_argument('--no-ssh-tunnel',
                                  help='Disable ssh tunnelling. Prevents program upload.',
                                  action='store_true')
        parsers[key].add_argument('--ssh-user', metavar="NAME", type=str,
                                  help="username to use as remote SSH name",
                                  default=str(config.ssh_user))

    parsers['audit'].set_defaults(func=_audit)
    parsers['web'].set_defaults(func=_start_web)
    parsers['add'].set_defaults(func=_add_program)
    parsers['delete'].set_defaults(func=_delete_program)
    parsers['query'].set_defaults(func=_make_query)

    # parse the arguments

    return ap.parse_args()

def _get_host_and_port(url):
        """Given a URL as http://host:port, returns (host, port)."""
        parsed = parse.urlparse(url)
        return (parsed.hostname, parsed.port)


def main():
    args = parse_arguments()
    remotehost, remoteport = _get_host_and_port(args.remote_neo4j)
    no_ssh_tunnel = args.no_ssh_tunnel
    connection = None

    try:
        conn_args = (remotehost, remoteport, no_ssh_tunnel)
        with db_api.SextantConnection(*conn_args) as connection:
            args.func(connection, args)
    except sshmanager.SSHConnectionError as e:
        print(e.message)

    
if __name__ == '__main__':
    main()


