import logging

"""
Provide a class for writing to row-limited csv files.
"""
__all__ = ('CSVWriter',)


class CSVWriter(object):
    """
    Write to csv files, automatically opening new ones at row maximum.

    Provides a write(*args) method which will add a row to the currently open
    csv file (internally managed) if there is room in it, otherwise close it,
    silently open a new one and write to that.

    Attributes:
        base_path:
            The base path of the output files - which will have a full path
            of form "<base_path><number>.csv"
        headers:
            A list or tuple of strings which will be used as the column
            headers. Attempts to write a row of data will induce a check
            that the length of the data provided is exactly that of this
            argument.
        max_rows:
            The maximum number of rows to write in each file (including the
            header row) before opening a new file.

        _fmt:
            The format string which will be used to write a row to the csv
            file. Of form '{},{},...,{}\n'.
        _file:
            The currently open file.
        _file_count:
            The number of files that the CSVWriter has written to. The next
            file to be opened will have name '<base_path><_file_count>.csv'
        _row_count:
            The number of rows (including the header row) in the current file.
        _total_row_count:
            The number of rows (including the header rows) in ALL files.

    """
    # Filename fmt of output files - used with .format(base_path, number).
    _file_fmt = '{}{}.csv'

    def __init__(self, base_path, headers, max_rows):
        """
        Initialise the writer for writing.

        Arguments:
            base_path:
                The base path of the output files - which will have a full path
                of form "<base_path><number>.csv"
            headers:
                A list or tuple of strings which will be used as the column
                headers. Attempts to write a row of data will induce a check
                that the length of the data provided is exactly that of this
                argument.
            max_rows:
                The maximum number of rows to write in each file (including the
                header row) before opening a new file.
        """
        self.base_path = base_path
        self.headers = headers
        self.max_rows = max_rows

        self._fmt = ','.join('{}' for h in headers) + '\n'

        # The number of the file we are on and the line in it.
        self._file = None
        self._file_count = 0
        self._row_count = 0

        self._total_row_count = 0

        self._open_new_file()

    def _open_new_file(self):
        """
        Open a new file for editing, writing the headers in the first row.
        """
        self._close_file()

        path = CSVWriter._file_fmt.format(self.base_path, self._file_count)
        self._file = open(path, 'w+')
        self._file_count += 1
        self.write(*self.headers)

    def _close_file(self):
        """
        Close the current file.
        
        NOTE that this method should ALWAYS be called before attempting to read
        from the file as it ensures that all changes have been written to disk,
        not only buffered.
        """
        if self._file and not self._file.closed:
            logging.debug('csvwriter wrote {} lines to {}'
                          .format(self._row_count, self._file.name))
            self._file.close()

        self._row_count = 0

    def write(self, *args):
        """
        Add a row the to current file, or to a new one if max_rows is reached.

        The check against max_rows is made BEFORE writing the line.

        Raises:
            ValueError:
                If the length of *args is not exactly the length of
                self.headers - i.e. on attempt to write too many/too few items.

        Arguments:
            *args:
                Strings, which will be written into the columns of the current
                open csv file.
        """
        if not len(args) == len(self.headers):
            msg = 'Attempted to write {} entries to file {} with {} columns'
            raise ValueError(msg.format(len(args), self.base_path, 
                                        len(self.headers)))

        if self._row_count == self.max_rows:
            self._close_file()
            self._open_new_file()

        self._file.write(self._fmt.format(*args))
        self._row_count += 1
        self._total_row_count += 1

    def file_iter(self):
        """
        Return an iterator over the names of the files the writer has 
        written to.
        """
        fmt = CSVWriter._file_fmt
        return (fmt.format(self.base_path, i) for i in range(self._file_count))

    def finish(self):
        """
        Flush and close the current file. If a subsequent call to self.write
        is made, a new file will be created to contain it.

        Return the number of files we have written to and the total number
        of lines we have written.
        """
        self._close_file()
        return self._file_count, self._total_row_count 

