# -----------------------------------------
# Sextant
# Copyright 2014, Ensoft Ltd.
# Author: Patrick Stevens, using work from Patrick Stevens and James Harkin
# -----------------------------------------
# API to interact with a Neo4J server: upload, query and delete programs in a DB

from __future__ import print_function

__all__ = ("validate_query", "DBProgram", "FunctionQueryResult", "Function",
           "SextantConnection")

from sys import stdout

import re  # for validation of function/program names
import logging
from datetime import datetime
import os
import getpass
from collections import namedtuple
import random
import socket

import itertools
import subprocess
from time import time

import neo4jrestclient.client as neo4jrestclient

from sshmanager import SSHManager, SSHConnectionError 
from csvwriter import CSVWriter

# The directory on the local machine to which csv files will be written
# prior to copy over to the remote server.
TMP_DIR = '/tmp/sextant'

SEARCH_NAMES_LIMIT = 50

# A function is deemed 'common' if it has more than this
# many connections.
COMMON_CUTOFF = 10



def set_common_cutoff(common_def):
    """
    Sets the number of incoming connections at which we deem a function 'common'
    Default is 10 (which is used if this method is never called).
    :param common_def: number of incoming connections
    """
    global COMMON_CUTOFF
    COMMON_CUTOFF = common_def


def validate_query(string):
    """
    Checks whether we can allow a string to be passed into a Cypher query.
    :param string: the string we wish to validate
    :return: bool(the string is allowed)
    """
    regex = re.compile(r'^[A-Za-z0-9\-:\.\$_@\*\(\)%\+,]+$')
    return bool(regex.match(string))


class DBProgram(object):
    """
    Representation of a program in the database.

    Provides add_function and add_call methods which locally register functions
    and calls. The commit method uploads everything to the database.

    Attributes:
        uploader, uploader_id, program_name, date:
            As in __init__.
        
        _conn:
            The SextantConnection object managing the database connection.
        _ssh:
            The SSHManager object belonging to the SextantConnection.
        _db:
            The database object belonging to the SextantConnection.

        _tmp_dir:
            The user-specific location of the local temporary directory.

        func_writer:
            A CSVWriter object which manages the csv files containing the
            list of functions in the program.
        call_writer:
            A CSVWriter object which manages the csv files containing the
            list of function calls in the program.

        add_func_query:
            A string for the cypher query used to create functions from a csv
            file.
        add_call_query:
            A string for the cypher query used to create funciton calls from
            a csv file.
        add_program_query:
            A string for the cypher query used to create the program node.
    """

    def __init__(self, connection, program_name, uploader, uploader_id, date):
        """
        Initialise the database program.

        A local temporary folder is created at 'TMP_DIR-<user_name>'. 
        When functions or calls are added via the add_function/call methods, 
        they are registered in csv files which are stored in this directory.

        Committing the program copies these files to the neo4j server and
        cleans the local tmp folder.

        Raises:
            ValueError:
                If the program_name is not alphanumeric.
            CommandError:
                If the command to create the temporary directory failed.

        Arguments:
            connection:
                The SextantConnection object which manages the connection to
                the database.
            program_name:
                The name to register the program under in the database. Must be
                alphanumeric.
            uploader:
                The name of the user who uploaded the program.
            uploader_id:
                A numeric id of the user who uploaded the program.
            date:
                A string representing the upload date.
        """
        # Ensure an alphanumeric program name.
        if not validate_query(program_name):
            raise ValueError('program name must be alphanumeric, got: {}'
                             .format(program_name));

        self.uploader = uploader
        self.uploader_id = uploader_id

        self.program_name = program_name
        self.date = date
        
        self._conn = connection
        self._ssh = connection._ssh
        self._db = connection._db
        
        self._tmp_dir = '{}-{}'.format(TMP_DIR, getpass.getuser())

        # Make the local tmp file - csv files will be written into here.
        try:
            os.makedirs(self._tmp_dir)
        except OSError as e:
            if e.errno == os.errno.EEXIST: # File already exists.
                pass
            else:
                raise e

        
        tmp_path = os.path.join(self._tmp_dir, '{}_{{}}'.format(program_name))

        self.func_writer = CSVWriter(tmp_path.format('funcs'), 
                                     headers=['name', 'type'], 
                                     max_rows=5000)
        self.call_writer = CSVWriter(tmp_path.format('calls'), 
                                     headers=['caller', 'callee'], 
                                     max_rows=5000)

        # Define the queries we use to upload the functions and calls.
        self.add_func_query = (' USING PERIODIC COMMIT 250'
                 ' LOAD CSV WITH HEADERS FROM "file:{}" AS line'
                 ' WITH line, toInt(line.id) as lineid'
                 ' MATCH (n:program {{name: "{}"}})'
                 ' CREATE (n)-[:subject]->(m:func {{name: line.name,'
                 ' id: lineid, type: line.type}})')
        
        self.add_call_query = (' USING PERIODIC COMMIT 250'
                 ' LOAD CSV WITH HEADERS FROM "file:{}" AS line'
                 ' MATCH (p:program {{name: "{}"}})'
                 ' MATCH (p)-[:subject]->(n:func {{name: line.caller}})'
                 ' USING INDEX n:func(name)'
                 ' MATCH (p)-[:subject]->(m:func {{name: line.callee}})' 
                 ' USING INDEX m:func(name)'
                 ' CREATE (n)-[r:calls]->(m)')

        self.add_program_query = ('CREATE (p:program {{name: "{}", uploader: "{}", '
                ' uploader_id: "{}", date: "{}",'
                ' function_count: {}, call_count: {}}})')


    def __enter__(self):
        """
        Allow DBProgram to be used as  a context manager.
        """
        return self

    def __exit__(self, etype, evalue, etrace):
        """
        Make sure that all files are properly closed.
        """
        self.func_writer.finish()
        self.call_writer.finish()

        # Propagate the error if there is one.
        return False if etype is not None else True

    def add_function(self, name, typ='normal'):
        """
        Add a function.

        Arguments:
            name:
                The name of the function.
            typ:
                The type of the function, may be any string, but standard types
                are:
                    normal: we have the disassembly for this function
                    stub:   we have the name but not the disassembly - usually
                            an imported library function.
                    pointer: we know only that the function exists, not its
                            name or details.
        """
        self.func_writer.write(name, typ)

    def add_call(self, caller, callee):
        """
        Add a function call.

        Arguments:
            caller:
                The name of the function making the call.
            callee:
                The name of the function called.
        """
        self.call_writer.write(caller, callee)


    def _copy_local_to_remote_tmp_dir(self):
        """
        Move local tmp files to the server ready for upload.

        Return a tuple of iterators, the first over the paths on the remote
        machine of the function files, and the second over the paths of the
        call files.
        """
        print('Sending files to remote server...', end='')
        stdout.flush()
        remote_funcs = self._ssh.send_to_tmp_dir(self.func_writer.file_iter())
        remote_calls = self._ssh.send_to_tmp_dir(self.call_writer.file_iter())
        print('finished.')
        return remote_funcs, remote_calls

    def _clean_tmp_files(self, remote_paths):
        """
        Delete temporary files on the local and remote machine.

        Arguments:
            remote_paths:
                A list of the paths of the remote fils.
        """
        print('Cleaning temporary files...', end='')
        file_paths = list(itertools.chain(self.func_writer.file_iter(),
                                          self.call_writer.file_iter()))

        for path in file_paths: 
            os.remove(path)

        os.rmdir(self._tmp_dir)

        try:
            # If the parent sextant temp folder is empty, remove it.
            os.rmdir(TMP_DIR)
        except:
            # There is other stuff in TMP_DIR (i.e. from other users), so
            # leave it.
            pass

        self._ssh.remove_from_tmp_dir(remote_paths)

        print('done.')

    def _create_db_constraints(self):
        """
        Create indexes in the database on program and function names.

        The program name index is a constraint, which will also garuantee the
        uniqueness of program names.
        """
        # Prepare a transaction object which we use to execute cypher queries.
        tx = self._db.transaction(using_globals=False, for_query=True)

        tx.append('CREATE CONSTRAINT ON (p:program) ASSERT p.name IS UNIQUE')
        tx.append('CREATE INDEX ON :func(name)')

        # Apply the transaction.
        tx.commit()

    def commit(self):
        """
        Insert the program into the database.

        Move the local temp files created by our func_writer and call_writer
        to the database server's temp directory. From there use cypher queries
        to upload them into the database, before cleaning them up.
        """
        # Ensure that the most recent files are flushed and closed.
        func_file_count, func_line_count = self.func_writer.finish()
        call_file_count, call_line_count = self.call_writer.finish()

        # Account for the header line at the top of each file.
        func_count = func_line_count - func_file_count
        call_count = call_line_count - call_file_count
        
        # Get the remote path names as iterators, then make lists of them
        # so that we can iterate over them more than once.
        remote_f_iter, remote_c_iter = self._copy_local_to_remote_tmp_dir()
        remote_funcs, remote_calls = map(list, (remote_f_iter, remote_c_iter))

        # Create the indexes and constraints in the database.
        self._create_db_constraints()


        try:
            tx = self._db.transaction(using_globals=False, for_query=True)

            # Create the program node in the database.
            tx.append(self.add_program_query.format(self.program_name, self.uploader,
                                                    self.uploader_id, self.date,
                                                    func_count, call_count))
            tx.commit()

            # Create the functions.
            for files, query, descr in zip((remote_funcs, remote_calls),
                                           (self.add_func_query, self.add_call_query),
                                           ('funcs', 'calls')):
                start = time()
                for i, path in enumerate(files):
                    completed = int(100*float(i+1)/len(files))

                    print('\rUploading {}: {}%'.format(descr, completed), end='')
                    stdout.flush()

                    tx.append(query.format(path, self.program_name))
                    tx.commit()
                end = time()
                print(' done.')

        finally:
            # Cleanup temporary folders
            self._clean_tmp_files(remote_funcs + remote_calls)


class FunctionQueryResult:
    """A graph of function calls arising as the result of a Neo4J query."""

    def __init__(self, parent_db, program_name='', rest_output=None):
        self.program_name = program_name
        self._parent_db_connection = parent_db
        self.functions = self._rest_node_output_to_graph(rest_output)
        self._update_common_functions()

    def __eq__(self, other):
        # We make a dictionary so that we can perform easy comparison.
        selfdict = {func.name: func for func in self.functions}
        otherdict = {func.name: func for func in other.functions}

        return self.program_name == other.program_name and selfdict == otherdict

    def _update_common_functions(self):
        """
        Loop over all functions: increment the called-by count of their callees.
        """
        for func in self.functions:
            for called in func.functions_i_call:
                called.number_calling_me += 1

    def _rest_node_output_to_graph(self, rest_output):
        """
        Convert the output of a REST API query into our internal representation.
        :param rest_output: output of the REST call as a Neo4j QuerySequence
        :return: iterable of <Function>s ready to initialise self.functions.
        """

        if rest_output is None or not rest_output.elements:
            return []

        # How we store this is: a dict
        #   with keys  'functionname'
        #   and values [the function object we will use,
        #               and a set of (function names this function calls),
        #               and numeric ID of this node in the Neo4J database].

        result = {}

        # Initial pass for names of functions.

        # If the following assertion failed, we've probably called db.query
        # to get it to not return client.Node objects, which is wrong.
        # we attempt to handle this a bit later; this should never arise, but
        # we can cope with it happening in some cases, like the test suite.

        if type(rest_output.elements) is not list:
            logging.warning('Not a list: {}'.format(type(rest_output.elements)))

        for node_list in rest_output.elements:
            assert(isinstance(node_list, list))
            for node in node_list:
                if isinstance(node, neo4jrestclient.Node):
                    name = node.properties['name']
                    node_id = node.id
                    node_type = node.properties['type']
                else:  
                    # This is the handling we mentioned earlier;
                    # we are a dictionary instead of a list, as for some
                    # reason we've returned Raw rather than Node data.
                    # We should never reach this code, but just in case.
                    name = node['data']['name']
                    # hacky workaround to get the id
                    node_id = node['self'].split('/')[-1]
                    node_type = node['data']['type']

                result[name] = [Function(self.program_name,
                                         function_name=name,
                                         function_type=node_type),
                                set(),
                                node_id]

        # End initialisation of names-dictionary.

        if self._parent_db_connection is not None:
            # This is the normal case, of extracting results from a server.
            # We leave the other case in because it is useful for unit testing.

            # We collect the name-name pairs of caller-callee, batched for speed
            new_tx = self._parent_db_connection.transaction(using_globals=False,
                                                            for_query=True)
            for index in result:
                q = ("START n=node({})"
                     "MATCH n-[calls:calls]->(m)"
                     "RETURN n.name, m.name").format(result[index][2])
                new_tx.append(q)

            logging.debug('exec')
            results = new_tx.execute()

            # Results is a list of query results, each of those being a list of
            # calls.

            for call_list in results:
                if call_list:
                    # call_list has element 0 being an arbitrary call this
                    # function makes; element 0 of that call is the name of the
                    # function itself. Think {{'orig', 'b'}, {'orig', 'c'}}.
                    orig = call_list[0][0]
                    # result['orig'] is [<Function>, ('callee1','callee2')]
                    result[orig][1] |= set(list(zip(*call_list.elements))[1])
                    # recall: set union is denoted by |

        else:
            # We don't have a parent database connection.
            # This has probably arisen because we created this object from a
            # test suite, or something like that.
            for node in rest_output.elements:
                node_name = node[0].properties['name']
                result[node_name][1] |= {relationship.end.properties['name']
                                         for relationship in node[0].relationships.outgoing()}

        logging.debug('Relationships complete.')

        # named_function takes a function name and returns the Function object
        # with that name, or None if none exists.
        named_function = lambda name: result[name][0] if name in result else None

        for function, calls, node_id in result.values():
            what_i_call = [named_function(name)
                           for name in calls
                           if named_function(name) is not None]
            function.functions_i_call = what_i_call

        return [list_element[0]
                for list_element in result.values()
                if list_element[0]]

    def get_functions(self):
        """
        :return: a list of Function objects present in the query result
        """
        return self.functions

    def get_function(self, name):
        """
        Given a function name, returns the Function object which has that name.
        If no function with that name exists, returns None.
        """
        func_list = [func for func in self.functions if func.name == name]
        return None if len(func_list) == 0 else func_list[0]

class Function(object):
    """
    Represents a function which might appear in a FunctionQueryResult.
    """

    def __eq__(self, other):
        funcs_i_call_list = {func.name for func in self.functions_i_call}
        funcs_other_calls_list = {func.name for func in other.functions_i_call}

        return (self.parent_program == other.parent_program
                and self.name == other.name
                and funcs_i_call_list == funcs_other_calls_list
                and self.attributes == other.attributes)

    @property
    def number_calling_me(self):
        return self._number_calling_me

    @number_calling_me.setter
    def number_calling_me(self, value):
        self._number_calling_me = value
        self.is_common = (self._number_calling_me > COMMON_CUTOFF)

    def __init__(self, program_name='', function_name='', function_type=''):
        self.parent_program = program_name
        self.attributes = []
        self.type = function_type
        self.functions_i_call = []
        self.name = function_name
        self.is_common = False
        self._number_calling_me = 0
        # Care: _number_calling_me is not automatically updated, except by
        # any invocation of FunctionQueryResult._update_common_functions.


class SextantConnection(object):
    """
    RESTful connection to a remote database.
    It can be used to create/delete/query programs.
    """

    ProgramWithMetadata = namedtuple('ProgramWithMetadata',
                                     ['uploader', 'uploader_id',
                                      'program_name', 'date', 
                                      'number_of_funcs', 'number_of_calls'])

    @staticmethod
    def _is_localhost(host, port):
        """
        Checks whether a host is an alias to localhost.

        Raises socket.gaierror if the host was not found.
        """
        addr = socket.getaddrinfo(host, port)[0][4][0]
        return addr in ('127.0.0.1', '::1')

    @staticmethod
    def _is_port_used(port):
        """
        Checks with the OS to see whether a port is open.

        Beware: port is passed directly to the shell. Make sure it is an integer.
        We raise ValueError if it is not.
        :param port: integer port to check for openness
        :return: bool(port is in use)
        """
        result = False

        # We follow:
        # http://stackoverflow.com/questions/2838244/get-open-tcp-port-in-python
        if not (isinstance(port, int) and port > 0):
            raise ValueError('port {} must be a positive integer.'.format(port))

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.bind(('127.0.0.1', port))
        except socket.error as e:
            if e.errno == os.errno.EADDRINUSE:
                result = True
            else:
                raise

        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

        return result  # that is, the port is not used

    @staticmethod
    def _get_unused_port():
        """
        Returns a port number between 10000 and 50000 which is not currently open.
        """

        keep_going = True
        while keep_going:
            portnum = random.randint(10000, 50000)
            keep_going = SextantConnection._is_port_used(portnum)
        return portnum


    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, etrace):
        self.close()
        return False if etype is not None else True


    def __init__(self, remotehost, remoteport, no_ssh_tunnel=False):
        """
        Initialise the database and ssh connections.

        Arguments:
            remotehost:
                The remote host name to connect to.
            remoteport:
                The port number to connect to on the remote host.
            no_ssh_tunnel:
                Disables the SSHManager if True. Prevents program upload.
        """

        self.remote_host = remotehost
        self.remote_port = remoteport


        self._no_ssh_tunnel = no_ssh_tunnel
        self._ssh = None
        self._db = None

        self.open()

    def open(self):
        local_port = SextantConnection._get_unused_port()
        is_localhost = SextantConnection._is_localhost(self.remote_host, self.remote_port)

        if self._no_ssh_tunnel and not is_localhost:
            raise SSHConnectionError('Cannot connect to the remote database '
                                     'without an ssh connection.')
        else:
            # Either we are making an ssh tunnel or we are contacting localhost.
            self._ssh = SSHManager(local_port, 
                                   self.remote_host, 
                                   self.remote_port, 
                                   is_localhost=is_localhost)

            port = self.remote_port if is_localhost else local_port
            url = 'http://localhost:{}'.format(port)

        self._db = neo4jrestclient.GraphDatabase(url)
 
    def close(self):
        """
        Close the ssh connection to clean up its resources.
        """
        if self._ssh:
            self._ssh.close()

    def new_program(self, program_name):
        """
        Request that the remote database create a new program with the given name.
        This procedure will create a new program remotely; you can manipulate
          that program using the returned DBProgram object.
        The name can appear in the database already, but this is not recommended
          because then delete_program will not know which to delete. Check first
          using self.check_program_exists.
        The name specified must pass validate_query()ion; this is a measure
          to prevent Cypher injection attacks.
        :param program_name: string program name
        :return: DBProgram instance if successful
        """

        if not validate_query(program_name):
            raise ValueError("{} is not a valid program name"
                             .format(program_name))
        
        uploader = getpass.getuser()
        uploader_id = os.getuid()
        timestr = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

        return DBProgram(self, program_name, uploader, 
                         uploader_id, date=timestr)

    def delete_program(self, program_name):
        """
        Request that the remote database delete a specified program.
        :param program_name: a string which must be alphanumeric only
        :return: bool(request succeeded)
        """
        if not program_name in self.get_program_names():
            print('No program `{}` in the database'.format(program_name))
            return True
        else:
            print('Deleting `{}` from the database. '
                  'This may take some time for larger programs.'
                  .format(program_name))

        start = time()
        tx = self._db.transaction(using_globals=False, for_query=True)

        count_query = (' MATCH (p:program {{name: "{}"}})'
                       ' RETURN p.function_count, p.call_count'
                       .format(program_name))

        tx.append(count_query)
        func_count, call_count = tx.commit()[0].elements[0]

        del_call_query = ('OPTIONAL MATCH (p:program {{name: "{}"}})'
                          '-[:subject]->(f:func)-[c:calls]->()'
                          ' WITH c LIMIT 5000 DELETE c RETURN count(distinct(c))'
                          .format(program_name))

        del_func_query = ('OPTIONAL MATCH (p:program {{name: "{}"}})'
                          '-[s:subject]->(f:func)'
                          ' WITH s, f LIMIT 5000 DELETE s, f RETURN count(f)'
                          .format(program_name))

        del_prog_query = ('MATCH (p:program {{name: "{}"}}) DELETE p'
                          .format(program_name))

        # Delete calls first, a node may not be deleted until all relationships
        # referencing it are deleted.
        for count, query, descr in zip((call_count, func_count),
                                       (del_call_query, del_func_query),
                                       ('calls', 'funcs')): 
            # Change tracks whether the last delete did anything. We would
            # like to use: while done < count: ..., but if the program has
            # already been partially deleted then this will never terminate.
            # Furthermore, if there are no functions or no calls, the while
            # loop will be appropriately skipped.
            change = count
            done = 0
            while change:
                completed = int(100 * float(done)/count)
                print('\rDeleting {}: {}%'.format(descr, completed), end='')
                stdout.flush()

                tx.append(query)
                change = tx.commit()[0].elements[0][0]
                done += change
            if done:
                print(' done.')

        # Delete the program node.
        tx.append(del_prog_query)
        tx.commit()

        end = time()
        print('Finished in {:.2f}s.'.format(end - start))

        return True


    def _execute_query(self, prog_name, query):
        """
        Executes a Cypher query against the remote database.
        Note that this returns a FunctionQueryResult, so is unsuitable for any
          other expected outputs (such as lists of names). For those instances,
          it is better to run self._parent_database_connection_object.query
          explicitly.
        Intended only to be used for non-updating queries
          (such as "get functions" rather than "create").
        :param prog_name: name of the program the result object will reflect
        :param query: verbatim query we wish the server to execute
        :return: a FunctionQueryResult corresponding to the server's output
        """
        rest_output = self._db.query(query, returns=neo4jrestclient.Node)

        return FunctionQueryResult(parent_db=self._db,
                                   program_name=prog_name,
                                   rest_output=rest_output)

    def get_program_names(self):
        """
        Execute query to retrieve a list of all programs in the database.
        Any name in this list can be used verbatim in any SextantConnection
          method which requires a program-name input.
        :return: a list of function-name strings.
        """
        q = 'MATCH (n:program) RETURN n.name'
        program_names = self._db.query(q, returns=str).elements

        return set(el[0] for el in program_names)


    def programs_with_metadata(self):
        """
        Returns a set of namedtuples which represent the current database.
        
        The namedtuples have .uploader, .uploader_id, .program_name, .date,
        .number_of_funcs.
        :return: set of namedtuples
       
        """
        
        q = (' MATCH (p:program)'
             ' RETURN p.uploader, p.uploader_id, p.name, p.date,'
             ' p.function_count, p.call_count')
        result = self._db.query(q)
        return {self.ProgramWithMetadata(*res) for res in result}

    def check_program_exists(self, program_name):
        """
        Execute query to check whether a program with the given name exists.
        Returns False if the program_name fails validation (i.e. is possibly
        unsafe as a string in a cypher query).
        :return: bool(the program exists in the database).
        """

        if not validate_query(program_name):
            return False

        q = ('MATCH (p:program {{name: "{}"}}) RETURN p LIMIT 1'
             .format(program_name))

        result = self._db.query(q, returns=neo4jrestclient.Node)
        return bool(result)

    def check_function_exists(self, program_name, function_name):
        """
        Execute query to check whether a function with the given name exists.
        We only check for functions which are children of a program with the
          given program_name.
        :param program_name: string name of the program within which to check
        :param function_name: string name of the function to check for existence
        :return: bool(names validate correctly, and function exists in program)
        """
        if not validate_query(program_name):
            return False

        pmatch = '(:program {{name: "{}"}})'.format(program_name)
        fmatch = '(f:func {{name: "{}"}})'.format(function_name)
        # be explicit about index usage
        q = (' MATCH {}-[:subject]->{} USING INDEX f:func(name)'
             ' RETURN f LIMIT 1'.format(pmatch, fmatch))

        # result will be an empty list if the function was not found
        result = self._db.query(q, returns=neo4jrestclient.Node)
        return bool(result)

    def get_function_names(self, program_name, search, max_funcs):
        """
        Execute query to retrieve a list of all functions in the program.
        Any of the output names can be used verbatim in any SextantConnection
          method which requires a function-name input.
        :param program_name: name of the program whose functions to retrieve
        :return: None if program_name doesn't exist in the remote database,
          a set of function-name strings otherwise.
        """

        if not validate_query(program_name):
            return set()

        if not search:
            q = (' MATCH (:program {{name: "{}"}})-[:subject]->(f:func)'
                 ' RETURN f.name LIMIT {}').format(program_name, max_funcs)
        else:
            q = (' MATCH (:program {{name: "{}"}})-[:subject]->(f:func)'
                 ' WHERE f.name =~ ".*{}.*" RETURN f.name LIMIT {}'
                 .format(program_name, search, max_funcs))
        return {func[0] for func in self._db.query(q)}

    def get_all_functions_called(self, program_name, function_calling):
        """
        Execute query to find all functions called by a function (indirectly).
        If the given function is not present in the program, returns None;
          likewise if the program_name does not exist.
        :param program_name: a string name of the program we wish to query under
        :param function_calling: string name of a function whose children to find
        :return: FunctionQueryResult, maximal subgraph rooted at function_calling
        """

        if not self.check_function_exists(program_name, function_calling):
            return None

        q = (' MATCH (p:program {{name: "{}"}})-[:subject]->(f:func {{name: "{}"}})'
             ' USING INDEX f:func(name)'
             ' MATCH (f)-[:calls*]->(g) RETURN distinct f, g'
             .format(program_name, function_calling))

        return self._execute_query(program_name, q)

    def get_all_functions_calling(self, program_name, function_called):
        """
        Execute query to find all functions which call a function (indirectly).
        If the given function is not present in the program, returns None;
          likewise if the program_name does not exist.
        :param program_name: a string name of the program we wish to query
        :param function_called: string name of a function whose parents to find
        :return: FunctionQueryResult, maximal connected subgraph with leaf function_called
        """

        if not self.check_function_exists(program_name, function_called):
            return None

        q = (' MATCH (p:program {{name: "{}"}})-[:subject]->(g:func {{name: "{}"}})'
             ' USING INDEX g:func(name)'
             ' MATCH (f)-[:calls*]->(g) WHERE f.name <> "{}"'
             ' RETURN distinct f , g')
        q = q.format(program_name, function_called, program_name)

        return self._execute_query(program_name, q)

    def get_call_paths(self, program_name, function_calling, function_called):
        """
        Execute query to find all possible routes between two specific nodes.
        If the given functions are not present in the program, returns None;
          ditto if the program_name does not exist.
        :param program_name: string program name
        :param function_calling: string
        :param function_called: string
        :return: FunctionQueryResult, the union of all subgraphs reachable by
          adding a source at function_calling and a sink at function_called.
        """

        if not self.check_program_exists(program_name):
            return None

        if not self.check_function_exists(program_name, function_called):
            return None

        if not self.check_function_exists(program_name, function_calling):
            return None

        q = (' MATCH (p:program {{name: "{}"}})-[:subject]->(start:func {{name: "{}"}})'
             ' USING INDEX start:func(name)'
             ' MATCH (p)-[:subject]->(end:func {{name: "{}"}})'
             ' USING INDEX end:func(name)'
             ' MATCH path=(start)-[:calls*]->(end)'
             ' WITH DISTINCT nodes(path) AS result'
             ' UNWIND result AS answer'
             ' RETURN answer')
        q = q.format(program_name, function_calling, function_called)

        return self._execute_query(program_name, q)

    def get_whole_program(self, program_name):
        """Execute query to find the entire program with a given name.
        If the program is not present in the remote database, returns None.
        :param: program_name: a string name of the program we wish to return.
        :return: a FunctionQueryResult consisting of the program graph.
        """

        if not self.check_program_exists(program_name):
            return None

        q = (' MATCH (p:program {{name: "{}"}})-[:subject]->(f:func)'
             ' RETURN (f)'.format(program_name))
        return self._execute_query(program_name, q)

    def get_shortest_path_between_functions(self, program_name, func1, func2):
        """
        Execute query to get a single, shortest, path between two functions.
        :param program_name: string name of the program we wish to search under
        :param func1: the name of the originating function of our shortest path
        :param func2: the name of the function at which to terminate the path
        :return: FunctionQueryResult shortest path between func1 and func2.
        """
        if not self.check_program_exists(program_name):
            return None

        if not self.check_function_exists(program_name, func1):
            return None

        if not self.check_function_exists(program_name, func2):
            return None

        q = (' MATCH (p:program {{name: "{}"}})-[:subject]->(f:func {{name: "{}"}})'
             ' USING INDEX f:func(name)'
             ' MATCH (p)-[:subject]->(g:func {{name: "{}"}})'
             ' MATCH path=shortestPath((f)-[:calls*]->(g))'
             ' UNWIND nodes(path) AS ans'
             ' RETURN ans'.format(program_name, func1, func2))

        return self._execute_query(program_name, q)
