#!/usr/bin/python
import argparse
import subprocess
import logging

"""
Provide a parser class to extract functions and calls from an objdump file,
and a way to generate such a file from an object file.
"""
__all__ = ('Parser', 'run_objdump', 'FileNotFoundError')


class FileNotFoundError(Exception):
    """
    Exception raised when Parser fails to open its file.
    """
    pass


class Parser(object):
    """
    Extract functions and calls from an object file or an objdump output file.

    Only the specified sections of the disassembled code will be parsed.

    Attributes:
        path:
            Set to file_path in __init__.
        _file:
            Set to file_object in __init__.
        sections:
            Initialised by taking the sections argument to __init__ and
            and converting it to a set.
        ignore_ptrs:
            Set to ignore_ptrs in __init__.

        section_count:
            The number of sections that have been parsed.
        function_count:
            The number of functions that have been parsed.
        call_count:
            The number of function calls that have been parsed.
        function_ptr_count:
            The number of function pointers that have been detected.
        _known_stubs:
            A set of the names of functions with type 'stub' that have been
            parsed - used to avoid registering a stub multiple times.

    """
    def __init__(self, file_path, file_object=None, 
                 sections=None, ignore_ptrs=False,
                 add_function=None, add_call=None, 
                 started=None, finished=None):
        """
        Initialise the parser object.

        Raises:
            FileNotFoundError:
                If file_object was not provided and file_path couldn't be
                opened.

        Arguments:
            file_path:
                The path of the objdump output file to parse, or the path of an
                object file to run objdump on and then parse.
            file_object:
                None if file_path is the path to an object file.
                OR the file object (providing 'for line in file_object')
            sections:
                A list of the names of the disassembly sections to parse. An mepty
                list will result in all sections being parsed.
            ignore_ptrs:
                If True, calls to function pointers will be ignored during parsing.
            add_function:
                A function to call when a function is parsed. Takes:
                    name: name of the parsed function
                    type: type of the parsed function
            add_call:
                A function to call when a function call is passed. Takes:
                    caller: name of the calling function
                    callee: name of the called function
            started:
                A function to call when the parse begins. Takes:
                    parser: the Parser instance which has just began parsing..
            finished:
                A function to call when the parse completes. Takes:
                    parser: the Parser instance which has just finished parsing.
                e.g. if add_function/call have been set to write into files, 
                then finished may be set to properly flush and close them.
        """
        self.path = file_path
        try:
            self._file = file_object or self._open_file(file_path)
        except FileNotFoundError:
            raise

        self.sections = set(sections or [])
        self.ignore_ptrs = ignore_ptrs

        self.section_count = 0
        self.function_count = 0
        self.call_count = 0
        self.function_ptr_count = 0
        
        # Avoid adding duplicate function stubs (as these are detected from
        # function calls so may be repeated).
        self._known_stubs = set()

        # By default print information to stdout.
        def print_func(name, typ):
            print('func {:25}{}'.format(name, typ))

        def print_call(caller, callee):
            print('call {:25}{:25}'.format(caller, callee))

        def print_started(parser):
            print('parse started: {}[{}]'.format(self.path, ', '.join(self.sections)))


        def print_finished(parser):
            print('parsed {} functions and {} calls'.format(self.function_count, self.call_count))

        self.add_function = add_function or print_func
        self.add_call = add_call or print_call
        self.started = lambda: (started or print_started)(self)
        self.finished = lambda: (finished or print_finished)(self)


    def _get_function_ptr_name(self):
        """
        Return a name for a new function pointer.
        """
        name = 'func_ptr_{}'.format(self.function_ptr_count)
        self.function_ptr_count += 1
        return name

    def _add_function_normal(self, name):
        """
        Add a function which we have full assembly code for.
        """
        self.add_function(name, 'normal')
        self.function_count += 1

    def _add_function_ptr(self, name):
        """
        Add a function pointer.
        """
        self.add_function(name, 'pointer')
        self.function_count += 1

    def _add_function_stub(self, name):
        """
        Add a function stub - we have its name but none of its internals.
        """
        if not name in self._known_stubs:
            self._known_stubs.add(name)
            self.add_function(name, 'stub')
            self.function_count += 1

    def _add_call(self, caller, callee):
        """
        Add a function call from caller to callee.
        """
        self.add_call(caller, callee)
        self.call_count += 1

    def parse(self):
        """
        Parse self._file.
        """
        self.started()

        if self._file is not None:
            in_section = False          # if we are in one of self.sections
            current_function = None     # track the caller for function calls

            for line in self._file:
                if line.startswith('Disassembly'):
                    # 'Disassembly of section <name>:\n'
                    section = line.split(' ')[-1].rstrip(':\n')
                    in_section = section in self.sections if self.sections else True
                    if in_section:
                        self.section_count += 1

                elif in_section:
                    if line.endswith('>:\n'):
                        # '<address> <<function_identifier>>:\n'
                        # with <function_identifier> of form:
                        # <function_name>[@plt]
                        function_identifier = line.split('<')[-1].split('>')[0]

                        if '@' in function_identifier:
                            current_function = function_identifier.split('@')[0]
                            self._add_function_stub(current_function)
                        else:
                            current_function = function_identifier
                            self._add_function_normal(current_function)

                    elif 'call ' in line or 'callq ' in line:
                        # WHITESPACE to prevent picking up function names 
                        # containing 'call'

                        # '<hex>: <hex> [l]call [hex] <callee_info>\n'
                        callee_info = line.split(' ')[-1].rstrip('\n')

                        # Where <callee_info> is either
                        #  1) '*(<register>)'           call to a fn pointer
                        #  2) '$<hex>,$<hex>'           lcall to a fn pointer
                        #  3) '<<function_identifier>>' call to a named function
                        if '<' in callee_info and '>' in callee_info:
                            # call to a normal or stub function
                            # '<function_identifier>' is of form <name>[@/-/+]<...>
                            # from which we extract name
                            callee_is_ptr = False
                            function_identifier = callee_info.lstrip('<').rstrip('>\n')
                            if '@' in function_identifier:
                                callee = function_identifier.split('@')[0]
                                self._add_function_stub(callee)
                            else:
                                callee = function_identifier.split('-')[-1].split('+')[0]
                                # Do not add this fn now - it is a normal func
                                # so we know about it from elsewhere.

                        else:
                            # Some kind of function pointer call.
                            callee_is_ptr = True
                            if not self.ignore_ptrs:
                                callee = self._get_function_ptr_name()
                                self._add_function_ptr(callee)

                        # Add the call.
                        if not (self.ignore_ptrs and callee_is_ptr):
                            self._add_call(current_function, callee)
            
            self.finished()

            self._file.close()
            result = True
        else:
            result = False

        return result

    def _open_file(self, path):
        """
        Open and return the file at path.

        Raises:
            FileNotFoundError:
                If the file fails to open.

        Arguments:
            path:
                The path of the file to open.
        """
        try:
            result = open(path)
        except Exception as e:
            raise FileNotFoundError("parser failed to open `{}`: {}".format(path, e.strerror))

        return result


def run_objdump(input_file):
    """
    Run the objdump command on the file with the given path.

    Return the input file path and a file object representing the result of
    the objdump.

    Arguments:
        input_file:
            The path of the file to run objdump on.
        
    """
    # A single section can be specified for parsing with the -j flag,
    # but it is not obviously possible to parse multiple sections like this.
    p = subprocess.Popen(['objdump', '-d', input_file, '--no-show-raw-insn'], 
                         stdout=subprocess.PIPE)
    g = subprocess.Popen(['egrep', 'Disassembly|call(q)? |>:$'], stdin=p.stdout, stdout=subprocess.PIPE)
    return input_file, g.stdout


def main():
    """
    Run the parser from the command line.

    The path of the target file, the sections to view and the ignore function
    pointers flag are set with command line arguments.
    """
    ap = argparse.ArgumentParser(description="Parse the output of objdump.")
    ap.add_argument('--filepath', metavar="FILEPATH", 
                    help="path to input file", type=str, nargs=1)

    ap.add_argument('--sections-to-view', metavar="SECTIONS",
                    help="disassembly sections to view, eg '.text'; leave blank for 'all'",
                    type=str, nargs='*')
    ap.add_argument('--ignore-function-pointers', 
                    help='skip parsing calls to function pointers', 
                    action='store_true', default=False)

    args = ap.parse_args()
    
    filepath = args.filepath[0]
    sections = args.sections_to_view
    ignore_ptrs = args.ignore_function_pointers

    parser = Parser(filepath, sections, ignore_ptrs)
    parser.parse()


if __name__ == '__main__':
    main()
