import os
import getpass
import logging
import subprocess

"""Provide a class to manage an SSH tunnel and controller"""
__all__ = ('SSHConnectionError', 'SSHCommandError', 'SSHManager')

# The location of the temporary directory to create on the REMOTE machine.
# Temporary files will be scp'd here prior to upload to the neo4j database.
TMP_DIR = '/tmp/sextant'


class SSHConnectionError(Exception):
    """
    An exception raised when an attempt to establish an ssh conneciton fails.
    """
    pass


class SSHCommandError(Exception):
    """
    An exception raised when an attempt to run a command over ssh fails.
    """
    pass


class SSHManager(object):
    """
    Manage an ssh tunnel with port forwarding.

    Attributes:
        local_port:
            The port number on the local machine to forward.
        remote_host:
            The host to ssh into.
        remote_port:
            The port number on the remote host to connect to.
        ssh_user:
            The username to use for sshing - defaults to None, in which case
            the ssh connection uses the username of the user who ran sextant.

        _controller_name:
            The base of the identifying name for the ssh controller - the
            actual name will be a combination of this and the local port.
        _is_localhost:
            True if we are trying to ssh into localhost. In this case do not
            open the tunnel, just provide the right api so the rest of Sextant
            need not special case.
    """

    def __init__(self, local_port, remote_host, remote_port, 
                 ssh_user=None, is_localhost=False):
        """
        Open an SSH tunnel with multiplexing enabled.

        Raises:
            ValueError:
                If local_port or remote_port are not positive integers

        Arguments:
            local_port:
                The number of the local port to forward.
            remote_host:
                The name of the remote host to connect to.
            remote_port:
                The port number on the remote host to connect to.
            ssh_user:
                An alternative user name to use for the ssh login.
            is_localhost:
                True if we are trying to ssh into localhost.
        """
        if not (isinstance(local_port, int) and local_port > 0):
            raise ValueError(
                'Local port {} must be a positive integer.'.format(local_port))
        if not (isinstance(remote_port, int) and remote_port > 0):
            raise ValueError(
                'Remote port {} must be a positive integer.'.format(remote_port))

        self.local_port = local_port
        self.remote_host = remote_host
        self.remote_port = remote_port
        self.ssh_user = ssh_user

        self._tmp_dir = '{}-{}'.format(TMP_DIR, self.ssh_user or getpass.getuser())

        self._controller_name = 'sextantcontroller{}'.format(local_port)
        self._is_localhost = is_localhost

        self._open()

    def _open(self):
        """
        Helper function to open the SSH tunnel.

        Raises:
            SSHConnectionError:
                If the ssh command failed to run.
        """
        if self._is_localhost:
            return

        # This cmd string will be .format()ed in a few lines' time.
        cmd = ['ssh']

        if self.ssh_user:
            # ssh -l {user} ... sets the remote login username
            cmd.extend(['-l', self.ssh_user])

        # -L localport:localhost:remoteport forwards the port.
        port_fwd = '{}:localhost:{}'.format(self.local_port, self.remote_port)

        # -M makes SSH able to accept slave connections.
        # -S sets the location of a control socket (in this case, sextantcontroller.
        #    with a unique identifier appended, just in case we run sextant twice.
        #    simultaneously), so we know how to close the port again.
        # -f goes into background; -N does not execute a remote command;
        # -T says to remote host that we don't want a text shell.
        cmd.extend(['-M', '-S', self._controller_name, '-fNT', 
                    '-L', port_fwd, self.remote_host])

        logging.debug('Opening SSH tunnel with cmd: {}'.format(' '.join(cmd)))

        rc = subprocess.call(cmd)
        if rc:
            raise SSHConnectionError('SSH setup failed with error {}'.format(rc))

        logging.debug('SSH tunnel created')

        self._make_tmp_dir()

    def close(self):
        """
        Close the SSH tunnel after cleaning the temp directory.
        """
        if self._is_localhost:
            return

        # ssh -O sends a command to the slave specified in -S, -q for quiet.
        cmd = ['ssh', '-S', self._controller_name, 
               '-O', 'exit', '-q', self.remote_host]

        logging.debug('Shutting down SSH tunnel with cmd: `{}`'
                      .format(' '.join(cmd)))

        # SSH has a bug on some systems which causes it to ignore the -q flag
        # meaning it prints "Exit request sent." to stderr.
        # To avoid this, we grab stderr temporarily, and see if it's that string;
        # if it is, suppress it.
        pr = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = pr.communicate()
        if stderr.rstrip() != 'Exit request sent.':
            logging.error('SSH shutdown stderr: {}'.format(stderr))

        if pr.returncode == 0:
            logging.debug('Shut down successfully')
        else:
            logging.error('SSH shutdown failed with code {}'
                          .format(pr.returncode))

        # Clean the temporary directory we created earlier.
        self._delete_tmp_dir()

    def _call(self, *args):
        """
        Execute a command on the remote machine over SSH.

        Return a tuple of rc, stdout, stderr from the process call.

        Arguments:
            *args:
                Strings containing the individual words of the command to
                execute. E.g. _call('ls', '-lh', '.').
        """
        if self._is_localhost:
            return (1, None, 'Cannot call SSH command from localhost')

        ssh_cmd = ['ssh', '-S', self._controller_name, self.remote_host]
        ssh_cmd.extend(args)
        p = subprocess.Popen(ssh_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()

        if p.returncode:
            logging.debug('Call to `{}` failed with code: {}, stderr: {}'
                          .format(' '.join(ssh_cmd), p.returncode, stderr))

        return p.returncode, stdout, stderr

    def _make_tmp_dir(self):
        """
        Create the per-user temporary directory on the remote machine.
        """
        self._call('mkdir', '-p', self._tmp_dir)

    def _delete_tmp_dir(self):
        """
        Remove the temporary directory on the remote machine.
        """
        self._call('rm', '-r', self._tmp_dir)


    def send_to_tmp_dir(self, path_list):
        """
        Send the specified files to the temporary directory on the remote machine.

        Return an iterator of save paths on the remote machine.  
        Raises:
            ValueError:
                If no file paths were provided, or if one or more of the
                provided paths is not an actual file.
            SSHCommandError:
                If the scp command failed for any reason.

        Arguments:
            path_list:
                Iterator of paths to the files on the local machine. All files 
                will be checked before copying to ensure that they exist and 
                to prevent passing arbitrary arguments to the ssh _call 
                command.
        """
        if not path_list:
            raise ValueError('attempt to copy zero files')

        # If we are in localhost, we are not controlling the TMP_DIR,
        # so the files are already there.
        if self._is_localhost:
            return path_list

        # Make sure we can take the len of path_list and iterate over it
        # more than once.
        path_list = list(path_list)

        # Check that actual files are being copied - not random strings.
        to_copy = [f for f in path_list if os.path.isfile(f)]

        if len(to_copy) < len(path_list):
            missed = [f for f in path_list if not f in to_copy]
            raise ValueError('Attempted to copy non existant files: {}'
                             .format(', '.join(missed)))

        scp_cmd = ['scp']
        scp_cmd.extend(to_copy)
        scp_cmd.append('{}:{}'.format(self.remote_host, self._tmp_dir))

        proc = subprocess.Popen(scp_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        rc = proc.wait()
        if rc:
            raise SSHCommandError('scp failed with code {}: {}'.format(rc, stderr))

        return (os.path.join(self._tmp_dir, os.path.basename(f)) for f in to_copy)

    def remove_from_tmp_dir(self, path_list):
        """
        Delete the files specified as arguments from the remote machine.

        The output of send_to_tmp_dir may be passed as input to this function.

        Raises:
            SSHCommandError:
                If the rm command fails for any reason.

        Arguments:
            path_list:
                Iterator of paths of the files on the remote machine, relative 
                to the temporary directory. E.g. remove_from_tmp_dir('foo') 
                will delete the file self._tmp_dir/foo
        """
        if self._is_localhost:
            return

        # Assume we can trust this file list.
        paths = [os.path.join(self._tmp_dir, os.path.basename(f)) for f in path_list]
        self._call('rm', *paths)

            



