#!/usr/bin/python
import unittest
from csvwriter import CSVWriter
import subprocess
from os import listdir

class TestSequence(unittest.TestCase):
    def get_writer(self, path='tmp_test', headers=['name', 'type'], split=100):
        return CSVWriter(path, headers, split)

    def tearDown(self):
        to_rm = [f for f in listdir('.') if f.startswith('tmp_test') and f.endswith('.csv')]
        if to_rm:
            rc = subprocess.call(['rm'] + to_rm)
            if rc:
                msg = 'failed to clean'
            else:
                msg = 'cleaned'
            print('{} {} files {}'.format(msg, len(to_rm), to_rm))

    def test_headers(self):
        # check that headers are being written correctly
        headers = ['some', 'headers', 'to', 'check']
        writer = self.get_writer(headers=headers)
        writer.finish()

        expected_path = 'tmp_test0.csv'
        self.assertEquals(writer.file_iter().next(), expected_path)
        writer_file = open('tmp_test0.csv', 'r')

        self.assertEquals(writer_file.readline(), 'some,headers,to,check\n')
        self.assertFalse(writer_file.readline()) # check that nothing extra is written

        writer_file.close()

    def test_writing(self):
        # check that csv entries are written correctly, and errors 
        # appropriately raised for invalid input
        writer = self.get_writer()

        self.assertRaises(ValueError, writer.write, 'too short')
        self.assertRaises(ValueError, writer.write, 'slightly', 'too', 'long')
        writer.write('just', 'write')

        writer.finish()

        writer_file = open(writer.file_iter().next(), 'r+')

        self.assertEqual(writer_file.readline(), 'name,type\n')
        self.assertEqual(writer_file.readline(), 'just,write\n')
        self.assertFalse(writer_file.readline())

        writer_file.close()

    def test_split(self):
        split = 10
        files = 10
        writer = self.get_writer(split=split)

        for i in xrange(files*(split-1)): # split-1 to account for header line
            writer.write('an', 'entry')

        writer.finish()

        gen_count = sum(1 for f in writer.file_iter())
        self.assertEqual(gen_count, files,
                         'generated {} files, expected {}'
                         .format(gen_count, files))

        for f in writer.file_iter():
            with open(f, 'r+') as wf:
                header_line = wf.readline()
                header_expected = 'name,type\n'
                self.assertEqual(header_line, header_expected,
                                 '{} contained header {}, expected {}'
                                 .format(f, header_line, header_expected)) # check headers

            # check line count
            with open(f, 'r+') as wf:
                line_count = sum(1 for line in wf)
                self.assertEqual(line_count, split,
                                 '{} contained {} lines, expected {}'
                                 .format(f, line_count, split))


if __name__ == '__main__':
    unittest.main()


