#!/usr/bin/python
from collections import defaultdict
import unittest
import subprocess

import objdump_parser as parser

DUMP_FILE = 'test_resources/parser_test.dump'

class TestSequence(unittest.TestCase):
    def setUp(self):
        pass
    
    def add_function(self, dct, name, typ):
        self.assertFalse(name in dct, "duplicate function added: {} into {}".format(name, dct.keys()))
        dct[name] = typ

    def add_call(self, dct, caller, callee):
        dct[caller].append(callee)

    def do_parse(self, path=DUMP_FILE, sections=['.text'], ignore_ptrs=False):
        functions = {}
        calls = defaultdict(list)
        
        # set the Parser to put output in local dictionaries
        add_function = lambda n, t: self.add_function(functions, n, t)
        add_call = lambda a, b: self.add_call(calls, a, b)

        p = parser.Parser(path, sections=sections, ignore_ptrs=ignore_ptrs, 
                          add_function=add_function, add_call=add_call)
        res = p.parse()

        parser.add_function = None
        parser.add_call = None

        return res, functions, calls


    def test_open(self):
        self.assertRaises(parser.FileNotFoundError, parser.Parser, file_path='rubbish file')

    def test_functions(self):
        # ensure that the correct functions are listed with the correct types
        res, funcs, calls = self.do_parse()

        for name, typ in zip(['normal', 'duplicates', 'wierd$name', 'printf', 'func_ptr_3'], 
                             ['normal', 'normal', 'normal', 'stub', 'pointer']):
            self.assertTrue(name in funcs, "'{}' not found in function dictionary".format(name))
            self.assertEquals(funcs[name], typ)

        self.assertFalse('__gmon_start__' in funcs, "don't see a function defined in .plt")

    def test_no_ptrs(self):
        # ensure that the ignore_ptrs flags is working
        res, funcs, calls = self.do_parse(ignore_ptrs=True)

        self.assertFalse('pointer' in funcs.values())
        self.assertEqual(len(calls['normal']), 2)


    def test_calls(self):
        res, funcs, calls = self.do_parse()

        self.assertTrue('normal' in calls['main'])
        self.assertTrue('duplicates' in calls['main'])

        normal_calls = sorted(['wierd$name', 'printf', 'func_ptr_3'])
        self.assertEquals(sorted(calls['normal']), normal_calls)

        self.assertEquals(calls['duplicates'].count('normal'), 2)
        self.assertEquals(calls['duplicates'].count('printf'), 2, 
                          "expected 2 printf calls in {}".format(calls['duplicates']))
        self.assertTrue('func_ptr_4' in calls['duplicates'])
        self.assertTrue('func_ptr_5' in calls['duplicates'])

    def test_sections(self):
        res, funcs, calls = self.do_parse(sections=['.plt', '.text'])

        # check that we have got rid of the @s in the names
        self.assertTrue('@' not in ''.join(funcs.keys()), "check names are extracted correctly")
        self.assertTrue('__gmon_start__' in funcs, "see a function defined only in .plt")

    
if __name__ == '__main__':
    unittest.main()
