# -----------------------------------------
# Sextant
# Copyright 2014, Ensoft Ltd.
# Author: Patrick Stevens, using work from Patrick Stevens and James Harkin
# -----------------------------------------
# Given a program file to upload, or a program name to delete from the server, does the right thing.

from __future__ import print_function

__all__ = ("upload_program", "delete_program")

from .db_api import SextantConnection
from .sshmanager import SSHConnectionError
from .objdump_parser import Parser, run_objdump
from os import path
from time import time
import subprocess
import sys

import logging

def upload_program(connection, user_name, file_path, program_name=None, 
                   not_object_file=False):
    """
    Upload a program's functions and call graph to the database.

    Arguments:
        connection:
            The SextantConnection object that manages the database connection.
        user_name:
            The user name of the user uploading the program.
        file_path:
            The path to either: the output of objdump (if not_object_file is 
            True) OR to a binary file if (not_object_file is False).
        program_name:
            An optional name to give the program in the database, if not
            specified then <user_name>-<file name> will be used.
        not_object_file:
            Flag controlling whether file_path is pointing to a dump file or
            a binary file.
    """
    if not connection._ssh:
        raise SSHConnectionError('An SSH connection is required for '
                                 'program upload.')

    if not program_name:
        file_no_ext = path.basename(file_path).split('.')[0]
        program_name = '{}-{}'.format(user_name, file_no_ext)


    if program_name in connection.get_program_names():
        raise ValueError('A program with name `{}` already exists in the database'
                         .format(program_name))


    print('Uploading `{}` to the database. '
          'This may take some time for larger programs.'
          .format(program_name))
    start = time()

    if not not_object_file:
        print('Generating dump file...', end='')
        sys.stdout.flush()
        file_path, file_object = run_objdump(file_path)
        print('done.')
    else:
        file_object = None

    # Make parser and wire to DBprogram.
    with connection.new_program(program_name) as program:

        def start_parser(program):
            print('Parsing dump file...', end='')
            sys.stdout.flush()

        def finish_parser(parser, program):
            # Callback to make sure the program's csv files are flushed when
            # the parser completes.
            program.func_writer.finish()
            program.call_writer.finish()

            print('done: {} functions and {} calls.'
                  .format(parser.function_count, parser.call_count))

        parser = Parser(file_path = file_path, file_object = file_object,
                        sections=[],
                        add_function = program.add_function,
                        add_call = program.add_call,
                        started=lambda parser: start_parser(program),
                        finished=lambda parser: finish_parser(parser, program))
        parser.parse()

        program.commit()

    end = time()
    print('Finished in {:.2f}s.'.format(end-start))


def delete_program(connection, program_name):
    """
    Remove the specified program from the database.

    Arguments:
        connection:
            The SextantConnection object managing the database connection.
        program_name:
            The name of the program to remove from the database.
    """
    connection.delete_program(program_name)

