#!/usr/bin/python2
# -----------------------------------------
# Sextant
# Copyright 2014, Ensoft Ltd.
# Author: Patrick Stevens, James Harkin
# -----------------------------------------
# Note: this must be run in Python 2.

from twisted.web.server import Site, NOT_DONE_YET
from twisted.web.resource import Resource
from twisted.web.static import File
from twisted.internet import reactor
from twisted.internet.threads import deferToThread
from twisted.internet import defer

import logging
import os
import json
import requests

import sextant.db_api as db_api
import sextant.export as export
import sextant.environment as environment
import tempfile
import subprocess

from cgi import escape  # deprecated in Python 3 in favour of html.escape, but we're stuck on Python 2

# global SextantConnection object which deals with the port forwarding
CONNECTION = None
CONNECTION_MAX_FUNCTION_NAMES = 500

RESPONSE_CODE_OK = 200
RESPONSE_CODE_BAD_REQUEST = 400
RESPONSE_CODE_NOT_FOUND = 404
RESPONSE_CODE_BAD_GATEWAY = 502
RESPONSE_CODE_NO_CONTENT = 204

def set_timeout(deferred, timeout=5.0):
    # cancel the given deferred instance after timeout seconds

    timeout = reactor.callLater(timeout, deferred.cancel)

    def got_result(result):
        # if the deferred returns before the timeout calls, cancel the timeout
        # and return the result of the deferred
        if timeout.active():
            timeout.cancel()

        return result

    return deferred.addBoth(got_result)

def defer_to_thread_with_timeout(timeout, function, *args, **kwargs):
    result = deferToThread(function, *args, **kwargs)
    return set_timeout(result, timeout)


class Echoer(Resource):
    # designed to take one name argument

    def render_GET(self, request):
        if "name" not in request.args:
            return '<html><body>Greetings, unnamed stranger.</body></html>'

        arg = escape(request.args["name"][0])
        return '<html><body>Hello %s!</body></html>' % arg


class SVGRenderer(Resource):

    @staticmethod
    def get_plot(program, suppress_common_functions=False, remove_self_calls=False):
        graph_dot = export.ProgramConverter.to_dot(program, suppress_common_functions,
                                                   remove_self_calls=remove_self_calls)

        file_written_to = tempfile.NamedTemporaryFile(delete=False)
        file_out = tempfile.NamedTemporaryFile(delete=False)
        file_written_to.write(graph_dot)
        file_written_to.close()
        subprocess.call(['dot', '-Tsvg', '-Kdot', '-o', file_out.name,
                         file_written_to.name])

        output = file_out.read().encode()
        file_out.close()
        return output

    @defer.inlineCallbacks
    def _render_plot(self, request):
        # the items in the args dict are lists - so use .get()[0] to retrieve
        args = request.args

        res_code = RESPONSE_CODE_OK 
        res_msg = None # set this in the logic

        #
        # Check if provided program name exists
        #

        name = args.get('program_name', [None])[0]

        if "program_name" is None:
            res_code = RESPONSE_CODE_BAD_REQUEST
            res_msg = "Supply 'program_name' parameter."

        if res_code is RESPONSE_CODE_OK:
            exists = yield deferToThread(CONNECTION.check_program_exists, name)
            if not exists:
                res_code = RESPONSE_CODE_NOT_FOUND
                res_fmt = "Program {} not found in database."
                res_msg = res_fmt.format(escape(name))

        #
        # We have a connection and a valid program - now setup the query
        #

        # We store query info as: 
        # <query_name>: (<function>, (<known args>), (<kwargs>)
        # where <known args> are local variables and <kwargs> are the keys we
        # look for in request.args, both tuples
        queries = {
            'whole_program': (
                CONNECTION.get_whole_program,
                ()
            ),
            'functions_calling': (
                CONNECTION.get_all_functions_calling,
                ('func1',)
            ),
            'functions_called_by': (
                CONNECTION.get_all_functions_called,
                ('func1',)
            ),
            'all_call_paths': (
                CONNECTION.get_call_paths,
                ('func1', 'func2')
            ),
            'shortest_call_path': (
                CONNECTION.get_shortest_path_between_functions,
                ('func1', 'func2')
            )
        }

        # default to whole_program if query is in request.args but has no value
        query_name = args.get('query', ['whole_program'])[0]

        # check for an invalid query
        query = queries.get(query_name, None)

        if query is None:
            # we do not have a query function for this query name
            res_code = RESPONSE_CODE_BAD_REQUEST
            res_fmt = "Invalid 'query' parameter, allowed: {}"
            res_msg = res_fmt.format(queries.keys())

        # extract any required keyword arguments from request.args
        if res_code is RESPONSE_CODE_OK:
            fn, kwargs = query
            
            # all args will be strings - use None to indicate missing argument
            req_args = tuple(args.get(kwarg, [None])[0] for kwarg in kwargs)
            missing_args = [kwarg for (kwarg, req_arg) in zip(kwargs, req_args)
                                if req_arg is None]

            if missing_args:
                # missing a kwarg from request.args
                res_code = RESPONSE_CODE_BAD_REQUEST 
                res_fmt = "Missing arguments {} to {}."
                res_msg = res_fmt.format(', '.join(missing_args), name)

        # if we are okay here we have a valid query with all required arguments
        if res_code is RESPONSE_CODE_OK:
            try:
                program = yield defer_to_thread_with_timeout(render_timeout, fn,
                                                             name, *req_args)
            except defer.CancelledError:
                # the timeout has fired and cancelled the request
                res_code = RESPONSE_CODE_BAD_REQUEST
                res_fmt = "The request timed out after {} seconds."
                res_msg = res_fmt.format(render_timeout)
            
        if res_code is RESPONSE_CODE_OK:
            # we have received a response to our request
            if program is None:
                res_code = RESPONSE_CODE_NOT_FOUND
                res_fmt = ("At least one of the input functions '{}' was not "
                           "found in program {}.")
                res_msg = res_fmt.format(', '.join(req_args), escape(name))
            elif not program.functions:
                res_code = RESPONSE_CODE_NO_CONTENT 
                res_fmt = ("The program {} is in the database but has no "
                           "functions.")
                res_msg = res_fmt.format(name)

        if res_code is RESPONSE_CODE_OK:
            # True if 'null' or 'true', otherwise False
            suppress_common_arg = args.get('suppress_common', ['false'])[0]
            suppress_common = suppress_common_arg in ('null', 'true')

            # we have a non-empty return - render it
            res_msg = yield deferToThread(self.get_plot, program, 
                                          suppress_common, 
                                          remove_self_calls=False)
            request.setHeader('content-type', 'image/svg+xml')

        request.setResponseCode(res_code)
        request.write(res_msg)
        request.finish()

    def render_GET(self, request):
        self._render_plot(request)
        return NOT_DONE_YET


class GraphProperties(Resource):

    @staticmethod
    def _get_program_names():
        return CONNECTION.get_program_names()

    @staticmethod
    def _get_function_names(program_name, search=""):
        max_funcs = 76;
        programs = CONNECTION.programs_with_metadata()
        result = CONNECTION.get_function_names(program_name, search, max_funcs)
        return result if len(result) < max_funcs else set()
        

    @defer.inlineCallbacks
    def _render_GET(self, request):
        if "query" not in request.args:
            request.setResponseCode(400)
            request.setHeader("content-type", "text/plain")
            request.write("Supply 'query' parameter of 'programs' or 'functions'.")
            request.finish()
            defer.returnValue(None)

        query = request.args['query'][0]

        if query == 'programs':
            request.setHeader("content-type", "application/json")
            prognames = yield deferToThread(self._get_program_names)
            request.write(json.dumps(list(prognames)))
            request.finish()
            defer.returnValue(None)

        elif query == 'functions':
            if "program_name" not in request.args:
                request.setResponseCode(400)
                request.setHeader("content-type", "text/plain")
                request.write("Supply 'program_name' parameter to ?query=functions.")
                request.finish()
                defer.returnValue(None)
            program_name = request.args['program_name'][0]
            search = request.args.get('search', [''])[0]

            funcnames = yield deferToThread(self._get_function_names, program_name, search)
            if funcnames is None:
                request.setResponseCode(404)
                request.setHeader("content-type", "text/plain")
                request.write("No program with name %s was found in the Sextant." % escape(program_name))
                request.finish()
                defer.returnValue(None)
            else:
                request.setHeader("content-type", "application/json")
                request.write(json.dumps(list(funcnames)))

                request.finish()
                defer.returnValue(None)

        else:
            request.setResponseCode(400)
            request.setHeader("content-type", "text/plain")
            request.write("'Query' parameter should be 'programs' or 'functions'.")
            request.finish()
            defer.returnValue(None)

    def render_GET(self, request):
        self._render_GET(request)
        return NOT_DONE_YET


def serve_site(connection, port):
    global CONNECTION

    CONNECTION = connection


    # serve static directory at root
    root = File(os.path.join(environment.RESOURCES_DIR, 'sextant', 'web'))

    # serve a dynamic Echoer webpage at /echoer.html
    root.putChild("echoer.html", Echoer())

    # serve a dynamic webpage at /Sextant_properties to return graph properties
    root.putChild("database_properties", GraphProperties())

    # serve a generated SVG at /output_graph.svg
    root.putChild('output_graph.svg', SVGRenderer())

    factory = Site(root)
    reactor.listenTCP(port, factory, interface=environment.SERVER_BIND)
    reactor.run()

if __name__ == '__main__':
    serve_site()
